@@ -61,11 +61,25 @@ hyperv_opts = [
                      "pending tasks. In some situations, the destroy "
                      "operation will fail (e.g. due to file locks), "
                      "requiring subsequent retries."),
+    cfg.BoolOpt('move_disks_on_cold_migration',
+                default=True,
+                help="Move the instance files to the instance dir configured "
+                     "on the destination host. You may consider disabling "
+                     "this when using multiple CSVs or shares and you wish "
+                     "the source location to be preserved."),
+]
+
+coordination_opts = [
+    cfg.StrOpt('backend_url',
+               default='file:///C:/OpenStack/Lock',
+               help='The backend URL to use for distributed coordination.'),
 ]
 
 CONF = nova.conf.CONF
+CONF.register_opts(coordination_opts, 'coordination')
 CONF.register_opts(hyperv_opts, 'hyperv')
 
 
 def list_opts():
-    return [('hyperv', hyperv_opts)]
+    return [('coordination', coordination_opts),
+            ('hyperv', hyperv_opts)]
