@@ -30,6 +30,7 @@ from os_win import utilsfactory
 from oslo_log import log as logging
 import six
 
+from compute_hyperv.nova import coordination
 from compute_hyperv.nova import eventhandler
 from compute_hyperv.nova import hostops
 from compute_hyperv.nova import imagecache
@@ -108,6 +109,8 @@ class HyperVDriver(driver.ComputeDriver):
         "supports_trusted_certs": True,
     }
 
+    use_coordination = False
+
     def __init__(self, virtapi):
         # check if the current version of Windows is supported before any
         # further driver initialisation.
@@ -126,6 +129,7 @@ class HyperVDriver(driver.ComputeDriver):
         self._imagecache = imagecache.ImageCache()
         self._image_api = image.API()
         self._pathutils = pathutils.PathUtils()
+        self._event_handler = eventhandler.InstanceEventHandler()
 
     def _check_minimum_windows_version(self):
         hostutils = utilsfactory.get_hostutils()
@@ -150,14 +154,23 @@ class HyperVDriver(driver.ComputeDriver):
         return False
 
     def init_host(self, host):
+        if self.use_coordination:
+            coordination.COORDINATOR.start()
+
         self._serialconsoleops.start_console_handlers()
-        event_handler = eventhandler.InstanceEventHandler(
-            state_change_callback=self.emit_event)
-        event_handler.start_listener()
+
+        self._set_event_handler_callbacks()
+        self._event_handler.start_listener()
 
         instance_path = self._pathutils.get_instances_dir()
         self._pathutils.check_create_dir(instance_path)
 
+    def _set_event_handler_callbacks(self):
+        # Subclasses may override this.
+        self._event_handler.add_callback(self.emit_event)
+        self._event_handler.add_callback(
+            self._vmops.instance_state_change_callback)
+
     def list_instance_uuids(self):
         return self._vmops.list_instance_uuids()
 
@@ -188,7 +201,7 @@ class HyperVDriver(driver.ComputeDriver):
         """Cleanup after instance being destroyed by Hypervisor."""
         self.unplug_vifs(instance, network_info)
 
-    def get_info(self, instance):
+    def get_info(self, instance, use_cache=True):
         return self._vmops.get_info(instance)
 
     def attach_volume(self, context, connection_info, instance, mountpoint,
@@ -208,7 +221,7 @@ class HyperVDriver(driver.ComputeDriver):
                                       instance,
                                       update_device_metadata=True)
 
-    def extend_volume(self, connection_info, instance):
+    def extend_volume(self, connection_info, instance, requested_size):
         self._volumeops.extend_volume(connection_info)
 
     def get_volume_connector(self, instance):
