@@ -24,12 +24,14 @@ import time
 
 from eventlet import timeout as etimeout
 from nova.api.metadata import base as instance_metadata
+from nova.compute import task_states
 from nova.compute import vm_states
 from nova import exception
 from nova import objects
 from nova.objects import fields
 from nova import utils
 from nova.virt import configdrive
+from nova.virt import event as virtevent
 from nova.virt import hardware
 from os_win import constants as os_win_const
 from os_win import exceptions as os_win_exc
@@ -140,11 +142,7 @@ class VMOps(object):
         info = self._vmutils.get_vm_summary_info(instance_name)
 
         state = constants.HYPERV_POWER_STATE[info['EnabledState']]
-        return hardware.InstanceInfo(state=state,
-                                     max_mem_kb=info['MemoryUsage'],
-                                     mem_kb=info['MemoryUsage'],
-                                     num_cpu=info['NumberOfProcessors'],
-                                     cpu_time_ns=info['UpTime'])
+        return hardware.InstanceInfo(state=state)
 
     def _create_root_device(self, context, instance, root_disk_info, vm_gen):
         path = None
@@ -159,7 +157,12 @@ class VMOps(object):
         root_iso_path_cached = self._imagecache.get_cached_image(context,
                                                                  instance)
         root_iso_path = self._pathutils.get_root_vhd_path(instance.name, 'iso')
-        self._pathutils.copyfile(root_iso_path_cached, root_iso_path)
+
+        if not os.path.exists(root_iso_path):
+            self._pathutils.copyfile(root_iso_path_cached, root_iso_path)
+        else:
+            LOG.info("Root iso '%s' already exists. Reusing it.",
+                     root_iso_path)
 
         return root_iso_path
 
@@ -174,6 +177,12 @@ class VMOps(object):
         root_vhd_path = self._pathutils.get_root_vhd_path(instance.name,
                                                           format_ext,
                                                           is_rescue_vhd)
+
+        if os.path.exists(root_vhd_path):
+            LOG.info("Root vhd '%s' already exists. Reusing it.",
+                     root_vhd_path)
+            return root_vhd_path
+
         root_vhd_size = instance.flavor.root_gb * units.Gi
 
         try:
@@ -238,8 +247,18 @@ class VMOps(object):
             self.create_ephemeral_disk(instance.name, eph)
 
     def create_ephemeral_disk(self, instance_name, eph_info):
-        self._vhdutils.create_dynamic_vhd(eph_info['path'],
-                                          eph_info['size'] * units.Gi)
+        if not os.path.exists(eph_info['path']):
+            self._vhdutils.create_dynamic_vhd(eph_info['path'],
+                                              eph_info['size'] * units.Gi)
+        else:
+            LOG.info("Ephemeral '%s' disk already exists. Reusing it.",
+                     eph_info['path'])
+
+    def get_attached_ephemeral_disks(self, instance_name):
+        vm_image_disks = self._vmutils.get_vm_storage_paths(
+            instance_name)[0]
+        return [image_path for image_path in vm_image_disks
+                if os.path.basename(image_path).lower().startswith('eph')]
 
     @staticmethod
     def _get_vif_metadata(context, instance_id):
@@ -256,20 +275,19 @@ class VMOps(object):
 
         return vif_metadata
 
-    def _save_device_metadata(self, context, instance, block_device_info):
+    def update_device_metadata(self, context, instance):
         """Builds a metadata object for instance devices, that maps the user
            provided tag to the hypervisor assigned device address.
         """
         metadata = []
 
         metadata.extend(self._get_vif_metadata(context, instance.uuid))
-        if block_device_info:
-            metadata.extend(self._block_dev_man.get_bdm_metadata(
-                context, instance, block_device_info))
+        metadata.extend(self._block_dev_man.get_bdm_metadata(context,
+                                                             instance))
 
-        if metadata:
-            instance.device_metadata = objects.InstanceDeviceMetadata(
-                devices=metadata)
+        instance.device_metadata = objects.InstanceDeviceMetadata(
+            devices=metadata)
+        instance.save()
 
     def set_boot_order(self, instance_name, vm_gen, block_device_info):
         boot_order = self._block_dev_man.get_boot_order(
@@ -290,11 +308,14 @@ class VMOps(object):
         if self._vmutils.vm_exists(instance_name):
             raise exception.InstanceExists(name=instance_name)
 
-        # Make sure we're starting with a clean slate.
-        self._delete_disk_files(instance)
-
         vm_gen = self.get_image_vm_generation(instance.uuid, image_meta)
 
+        instance_dir = self._pathutils.get_instance_dir(instance.name,
+                                                        create_dir=False)
+        if os.path.exists(instance_dir):
+            LOG.info("Instance directory already exists."
+                     "Reusing existing files.")
+
         self._block_dev_man.validate_and_update_bdi(
             instance, image_meta, vm_gen, block_device_info)
         root_device = block_device_info['root_disk']
@@ -309,7 +330,7 @@ class VMOps(object):
                 # This is supported starting from OVS version 2.5
                 self.plug_vifs(instance, network_info)
 
-            self._save_device_metadata(context, instance, block_device_info)
+            self.update_device_metadata(context, instance)
 
             if configdrive.required_by(instance):
                 configdrive_path = self._create_config_drive(context,
@@ -400,10 +421,10 @@ class VMOps(object):
         self.configure_remotefx(instance, vm_gen)
 
         self._vmutils.create_scsi_controller(instance_name)
-        self._attach_root_device(instance_name, root_device)
+        self._attach_root_device(context, instance, root_device)
         self.attach_ephemerals(instance_name, block_device_info['ephemerals'])
         self._volumeops.attach_volumes(
-            block_device_info['block_device_mapping'], instance_name)
+            context, block_device_info['block_device_mapping'], instance)
 
         serial_ports = self._get_image_serial_port_settings(image_meta)
         self._create_vm_com_port_pipes(instance, serial_ports)
@@ -414,8 +435,7 @@ class VMOps(object):
                                      vif['id'],
                                      vif['address'])
 
-        if CONF.hyperv.enable_instance_metrics_collection:
-            self._metricsutils.enable_vm_metrics_collection(instance_name)
+        self.configure_instance_metrics(instance_name)
 
         if secure_boot_enabled:
             certificate_required = self._requires_certificate(image_meta)
@@ -606,13 +626,14 @@ class VMOps(object):
             remotefx_max_resolution,
             vram_bytes)
 
-    def _attach_root_device(self, instance_name, root_dev_info):
+    def _attach_root_device(self, context, instance, root_dev_info):
         if root_dev_info['type'] == constants.VOLUME:
-            self._volumeops.attach_volume(root_dev_info['connection_info'],
-                                          instance_name,
+            self._volumeops.attach_volume(context,
+                                          root_dev_info['connection_info'],
+                                          instance,
                                           disk_bus=root_dev_info['disk_bus'])
         else:
-            self._attach_drive(instance_name, root_dev_info['path'],
+            self._attach_drive(instance.name, root_dev_info['path'],
                                root_dev_info['drive_addr'],
                                root_dev_info['ctrl_disk_addr'],
                                root_dev_info['disk_bus'],
@@ -629,6 +650,14 @@ class VMOps(object):
                     constants.BDI_DEVICE_TYPE_TO_DRIVE_TYPE[
                         eph['device_type']])
 
+                # This may be an ephemeral added by default by us, in which
+                # case there won't be a bdm object.
+                bdm_obj = getattr(eph, '_bdm_obj', None)
+                if bdm_obj:
+                    filename = os.path.basename(eph['path'])
+                    self._block_dev_man.update_bdm_connection_info(
+                        eph._bdm_obj, eph_filename=filename)
+
     def _attach_drive(self, instance_name, path, drive_addr, ctrl_disk_addr,
                       controller_type, drive_type=constants.DISK):
         if controller_type == constants.CTRL_TYPE_SCSI:
@@ -765,15 +794,16 @@ class VMOps(object):
         if not CONF.hyperv.config_drive_cdrom:
             configdrive_path = self._pathutils.get_configdrive_path(
                 instance.name, constants.DISK_FORMAT_VHD, rescue=rescue)
-            utils.execute(CONF.hyperv.qemu_img_cmd,
-                          'convert',
-                          '-f',
-                          'raw',
-                          '-O',
-                          'vpc',
-                          configdrive_path_iso,
-                          configdrive_path,
-                          attempts=1)
+            processutils.execute(
+                CONF.hyperv.qemu_img_cmd,
+                'convert',
+                '-f',
+                'raw',
+                '-O',
+                'vpc',
+                configdrive_path_iso,
+                configdrive_path,
+                attempts=1)
             self._pathutils.remove(configdrive_path_iso)
         else:
             configdrive_path = configdrive_path_iso
@@ -828,8 +858,26 @@ class VMOps(object):
             if backup_location:
                 self._pathutils.check_remove_dir(backup_location)
 
-    def destroy(self, instance, network_info, block_device_info,
-                destroy_disks=True, cleanup_migration_files=True):
+    def destroy(self, instance, *args, **kwargs):
+        # Nova allows destroying instances regardless of pending tasks.
+        # In some cases, we may not be able to properly delete instances
+        # while having a pending task (e.g. when snapshotting, due to file
+        # locks).
+        #
+        # We may append other locks for operations that are non preemptive.
+        # We should not rely on instance task states, which may be hanging.
+        @utils.synchronized(constants.SNAPSHOT_LOCK_TEMPLATE %
+                            dict(instance_uuid=instance.uuid))
+        def synchronized_destroy():
+            self._destroy(instance, *args, **kwargs)
+
+        if CONF.hyperv.force_destroy_instances:
+            self._destroy(instance, *args, **kwargs)
+        else:
+            synchronized_destroy()
+
+    def _destroy(self, instance, network_info, block_device_info,
+                 destroy_disks=True, cleanup_migration_files=True):
         instance_name = instance.name
         LOG.info("Got request to destroy instance", instance=instance)
 
@@ -838,6 +886,15 @@ class VMOps(object):
         instance_path = self._pathutils.get_instance_dir(instance.name,
                                                          create_dir=False)
 
+        # When reverting resizes, the manager will request instance files
+        # cleanup to be skipped if the hosts use shared storage, in which
+        # case we'd leak files if multiple CSVs are used. It's safe to
+        # cleanup the instance files when reverting resizes as long as we
+        # preserve the "*_revert" dir.
+        if instance.task_state == task_states.RESIZE_REVERTING:
+            destroy_disks = True
+            cleanup_migration_files = False
+
         try:
             if self._vmutils.vm_exists(instance_name):
 
@@ -1050,13 +1107,34 @@ class VMOps(object):
                 except Exception as exc:
                     LOG.exception("Failed to plug vif: '%s'.",
                                   vif, instance=instance)
-                    raise exception.VirtualInterfacePlugException(exc)
+                    raise exception.VirtualInterfacePlugException(
+                        six.text_type(exc))
 
     def unplug_vifs(self, instance, network_info):
         if network_info:
             for vif in network_info:
                 self._vif_driver.unplug(instance, vif)
 
+    def configure_instance_metrics(self, instance_name,
+                                   enable_network_metrics=False):
+        if not CONF.hyperv.enable_instance_metrics_collection:
+            LOG.debug("Instance metrics collection is not enabled.")
+            return
+
+        LOG.debug("Enabling instance %s metrics.", instance_name)
+        # Some metrics (cpu, memory and network) are disabled when the vms are
+        # powered off. This looks like a Hyper-V bug that we'll have to
+        # mitigate at the Nova driver level.
+        self._metricsutils.enable_vm_metrics_collection(instance_name)
+
+        # Network metrics are handled separately. The reason is that the vm
+        # must be running and the ports must be already attached in order to
+        # be able to enable those metrics.
+        if enable_network_metrics:
+            vif_ids = self._vmutils.get_vm_nic_names(instance_name)
+            for vif_id in vif_ids:
+                self._vif_driver.enable_metrics(instance_name, vif_id)
+
     def _get_image_serial_port_settings(self, image_meta):
         image_props = image_meta['properties']
         serial_ports = {}
@@ -1092,21 +1170,19 @@ class VMOps(object):
             return True
 
         if not self._hostutils.check_min_windows_version(10, 0):
-            # TODO(claudiub): add set log level to error after string freeze.
-            LOG.debug("vNIC hot plugging is supported only in newer "
+            LOG.error("vNIC hot plugging is supported only in newer "
                       "versions than Windows Hyper-V / Server 2012 R2.")
             return False
 
         if (self._vmutils.get_vm_generation(instance.name) ==
                 constants.VM_GEN_1):
-            # TODO(claudiub): add set log level to error after string freeze.
-            LOG.debug("Cannot hot plug vNIC to a first generation VM.",
+            LOG.error("Cannot hot plug vNIC to a first generation VM.",
                       instance=instance)
             return False
 
         return True
 
-    def attach_interface(self, instance, vif):
+    def attach_interface(self, context, instance, vif):
         if not self._check_hotplug_available(instance):
             raise exception.InterfaceAttachFailed(instance_uuid=instance.uuid)
 
@@ -1114,6 +1190,8 @@ class VMOps(object):
         self._vmutils.create_nic(instance.name, vif['id'], vif['address'])
         self._vif_driver.plug(instance, vif)
 
+        self.update_device_metadata(context, instance)
+
     def detach_interface(self, instance, vif):
         try:
             if not self._check_hotplug_available(instance):
@@ -1124,8 +1202,7 @@ class VMOps(object):
             self._vif_driver.unplug(instance, vif)
             self._vmutils.destroy_nic(instance.name, vif['id'])
         except os_win_exc.HyperVVMNotFoundException:
-            # TODO(claudiub): add set log level to error after string freeze.
-            LOG.debug("Instance not found during detach interface. It "
+            LOG.error("Instance not found during detach interface. It "
                       "might have been destroyed beforehand.",
                       instance=instance)
             raise exception.InterfaceDetachFailed(instance_uuid=instance.uuid)
@@ -1402,3 +1479,15 @@ class VMOps(object):
                           "its uuid. It may have been deleted meanwhile.",
                           instance_name)
                 ctxt.reraise = expect_existing
+
+    def instance_state_change_callback(self, event):
+        if event.transition in (virtevent.EVENT_LIFECYCLE_STARTED,
+                                virtevent.EVENT_LIFECYCLE_RESUMED):
+            # We can handle the following operations concurrently.
+            utils.spawn_n(self._serial_console_ops.start_console_handler,
+                          event.name)
+            utils.spawn_n(self.configure_instance_metrics,
+                          event.name,
+                          enable_network_metrics=True)
+        else:
+            self._serial_console_ops.stop_console_handler(event.name)
