@@ -18,8 +18,11 @@ import mock
 from nova.compute import power_state
 from nova.compute import task_states
 from nova.compute import vm_states
+from nova import exception
 from nova.network.neutronv2 import api as network_api
 from nova import objects
+from nova.virt import event as virtevent
+from os_win import constants as os_win_const
 from os_win import exceptions as os_win_exc
 
 from compute_hyperv.nova.cluster import clusterops
@@ -39,6 +42,7 @@ class ClusterOpsTestCase(test_base.HyperVBaseTestCase):
         network_api.API,
         clusterops.vmops.VMOps,
         clusterops.serialconsoleops.SerialConsoleOps,
+        clusterops.placement_utils.PlacementUtils,
     ]
 
     _FAKE_INSTANCE_NAME = 'fake_instance_name'
@@ -49,7 +53,10 @@ class ClusterOpsTestCase(test_base.HyperVBaseTestCase):
 
         self.clusterops = clusterops.ClusterOps()
         self.clusterops._context = self.context
+
         self._clustutils = self.clusterops._clustutils
+        self._network_api = self.clusterops._network_api
+        self._placement = self.clusterops._placement
 
     def test_get_instance_host(self):
         mock_instance = fake_instance.fake_instance_obj(self.context)
@@ -112,7 +119,7 @@ class ClusterOpsTestCase(test_base.HyperVBaseTestCase):
 
         spawn_args = mock_spawn.call_args_list[0][0]
         self.assertEqual(
-            self._clustutils.get_vm_owner_change_listener.return_value,
+            self._clustutils.get_vm_owner_change_listener_v2.return_value,
             spawn_args[0])
 
         cbk = spawn_args[1]
@@ -142,14 +149,15 @@ class ClusterOpsTestCase(test_base.HyperVBaseTestCase):
             mock_instance1.name, mock_instance1.host,
             self.clusterops._this_node)
 
+    @mock.patch.object(clusterops.ClusterOps, '_wait_for_pending_instance')
     @mock.patch.object(clusterops, 'LOG')
     @mock.patch.object(clusterops.ClusterOps, '_get_instance_by_name')
     def test_failover_migrate_no_instance(self, mock_get_instance_by_name,
-                                          mock_LOG):
+                                          mock_LOG,
+                                          mock_wait_pending_instance):
         mock_get_instance_by_name.return_value = None
 
         self.clusterops._failover_migrate(mock.sentinel.instance_name,
-                                          mock.sentinel.old_host,
                                           mock.sentinel.new_host)
 
         mock_LOG.debug.assert_called_once_with(
@@ -158,59 +166,74 @@ class ClusterOpsTestCase(test_base.HyperVBaseTestCase):
         self.assertFalse(
             self.clusterops._network_api.get_instance_nw_info.called)
 
+    @mock.patch.object(clusterops.ClusterOps, '_wait_for_pending_instance')
     @mock.patch.object(clusterops, 'LOG')
     @mock.patch.object(clusterops.ClusterOps, '_get_instance_by_name')
     def test_failover_migrate_migrating(self, mock_get_instance_by_name,
-                                        mock_LOG):
+                                        mock_LOG, mock_wait_pending_instance):
         instance = mock_get_instance_by_name.return_value
         instance.task_state = task_states.MIGRATING
 
         self.clusterops._failover_migrate(mock.sentinel.instance_name,
-                                          mock.sentinel.old_host,
-                                          mock.sentinel.new_host)
+                                          'new_host')
 
         mock_LOG.debug.assert_called_once_with(
             'Instance %s is live migrating.', mock.sentinel.instance_name)
 
+    @mock.patch.object(clusterops.ClusterOps, '_wait_for_pending_instance')
     @mock.patch.object(clusterops.ClusterOps, '_get_instance_by_name')
-    def test_failover_migrate_at_source_node(self, mock_get_instance_by_name):
+    def test_failover_migrate_at_source_node(self, mock_get_instance_by_name,
+                                             mock_wait_pending_instance):
         instance = mock_get_instance_by_name.return_value
-        old_host = 'old_host'
-        self.clusterops._this_node = old_host
+        instance.host = 'old_host'
+        self.clusterops._this_node = instance.host
 
         self.clusterops._failover_migrate(mock.sentinel.instance_name,
-                                          old_host, mock.sentinel.new_host)
+                                          'new_host')
 
         self.clusterops._vmops.unplug_vifs.assert_called_once_with(instance,
             self.clusterops._network_api.get_instance_nw_info.return_value)
 
+    @mock.patch.object(clusterops.ClusterOps, '_wait_for_pending_instance')
     @mock.patch.object(clusterops, 'LOG')
     @mock.patch.object(clusterops.ClusterOps, '_get_instance_by_name')
     def test_failover_migrate_not_this_node(self, mock_get_instance_by_name,
-                                            mock_LOG):
+                                            mock_LOG,
+                                            mock_wait_pending_instance):
         self.clusterops._this_node = 'new_host'
 
         self.clusterops._failover_migrate(mock.sentinel.instance_name,
-                                          None, 'host')
+                                          'host')
 
         mock_LOG.debug.assert_called_once_with(
             'Instance %s did not failover to this node.',
             mock.sentinel.instance_name)
 
+    @mock.patch.object(clusterops.ClusterOps, '_wait_for_pending_instance')
     @mock.patch.object(clusterops.ClusterOps, '_failover_migrate_networks')
     @mock.patch.object(clusterops.ClusterOps, '_nova_failover_server')
     @mock.patch.object(clusterops.ClusterOps, '_get_instance_by_name')
-    def test_failover_migrate_this_node(self, mock_get_instance_by_name,
-                                        mock_nova_failover_server,
-                                        mock_failover_migrate_networks):
+    def test_failover_migrate_changed_host(self, mock_get_instance_by_name,
+                                           mock_nova_failover_server,
+                                           mock_failover_migrate_networks,
+                                           mock_wait_pending_instance):
         instance = mock_get_instance_by_name.return_value
         old_host = 'old_host'
         new_host = 'new_host'
+        instance.host = old_host
         self.clusterops._this_node = new_host
+        self._clustutils.get_vm_host.return_value = new_host
+        # Placement exceptions shouldn't break the rest of the failover logic.
+        self._placement.move_compute_node_allocations.side_effect = (
+            exception.NovaException)
 
         self.clusterops._failover_migrate(mock.sentinel.instance_name,
-                                          old_host, new_host)
+                                          new_host)
 
+        mock_wait_pending_instance.assert_called_once_with(
+            mock.sentinel.instance_name)
+        self._clustutils.get_vm_host.assert_called_once_with(
+            mock.sentinel.instance_name)
         mock_get_instance_by_name.assert_called_once_with(
             mock.sentinel.instance_name)
         get_inst_nw_info = self.clusterops._network_api.get_instance_nw_info
@@ -219,11 +242,59 @@ class ClusterOpsTestCase(test_base.HyperVBaseTestCase):
         mock_nova_failover_server.assert_called_once_with(instance, new_host)
         mock_failover_migrate_networks.assert_called_once_with(
             instance, old_host)
+        self._placement.move_compute_node_allocations.assert_called_once_with(
+            self.clusterops._context, instance, old_host, new_host,
+            merge_existing=False)
         self.clusterops._vmops.plug_vifs.assert_called_once_with(
             instance, get_inst_nw_info.return_value)
         c_handler = self.clusterops._serial_console_ops.start_console_handler
         c_handler.assert_called_once_with(mock.sentinel.instance_name)
 
+    @mock.patch.object(clusterops.ClusterOps, '_wait_for_pending_instance')
+    @mock.patch.object(clusterops.ClusterOps, '_failover_migrate_networks')
+    @mock.patch.object(clusterops.ClusterOps, '_nova_failover_server')
+    @mock.patch.object(clusterops.ClusterOps, '_get_instance_by_name')
+    def test_failover_same_node(self, mock_get_instance_by_name,
+                                mock_nova_failover_server,
+                                mock_failover_migrate_networks,
+                                mock_wait_pending_instance):
+        # In some cases, the instances may bounce between hosts. We're testing
+        # the case in which the instance is actually returning to the initial
+        # host during the time in which we're processing events.
+        instance = mock_get_instance_by_name.return_value
+        old_host = 'old_host'
+        new_host = 'new_host'
+        instance.host = old_host
+        self.clusterops._this_node = old_host
+        self._clustutils.get_vm_host.return_value = old_host
+
+        self.clusterops._failover_migrate(mock.sentinel.instance_name,
+                                          new_host)
+
+        get_inst_nw_info = self.clusterops._network_api.get_instance_nw_info
+        get_inst_nw_info.assert_called_once_with(self.clusterops._context,
+                                                 instance)
+        mock_nova_failover_server.assert_called_once_with(instance, old_host)
+        self.clusterops._vmops.unplug_vifs.assert_not_called()
+        self.clusterops._vmops.plug_vifs.assert_called_once_with(
+            instance, get_inst_nw_info.return_value)
+        self._placement.move_compute_node_allocations.assert_not_called()
+        mock_failover_migrate_networks.assert_not_called()
+        c_handler = self.clusterops._serial_console_ops.start_console_handler
+        c_handler.assert_called_once_with(mock.sentinel.instance_name)
+
+    @mock.patch('time.sleep')
+    def test_wait_for_pending_instance(self, mock_sleep):
+        self._clustutils.get_cluster_group_state_info.side_effect = [
+            dict(state=os_win_const.CLUSTER_GROUP_PENDING),
+            dict(state=os_win_const.CLUSTER_GROUP_ONLINE)]
+
+        self.clusterops._wait_for_pending_instance(mock.sentinel.instance_name)
+
+        self._clustutils.get_cluster_group_state_info.assert_has_calls(
+            [mock.call(mock.sentinel.instance_name)] * 2)
+        mock_sleep.assert_called_once_with(2)
+
     def test_failover_migrate_networks(self):
         mock_instance = fake_instance.fake_instance_obj(self.context)
         fake_source = mock.MagicMock()
@@ -341,3 +412,17 @@ class ClusterOpsTestCase(test_base.HyperVBaseTestCase):
         self.assertEqual(mock.sentinel.host, mock_instance.host)
         self.assertEqual(mock.sentinel.host, mock_instance.node)
         mock_instance.save.assert_called_once_with(expected_task_state=[None])
+
+    @mock.patch.object(clusterops.ClusterOps, '_get_instance_by_name')
+    def test_instance_state_change_callback(self, mock_get_instance_by_name):
+        event = mock.Mock(transition=virtevent.EVENT_LIFECYCLE_STARTED)
+        mock_instance = mock_get_instance_by_name.return_value
+
+        self.clusterops.instance_state_change_callback(event)
+
+        mock_get_instance_by_name.assert_called_once_with(event.name)
+        self._network_api.get_instance_nw_info.assert_called_once_with(
+            self.context, mock_instance)
+        self.clusterops._vmops.plug_vifs.assert_called_once_with(
+            mock_instance,
+            self._network_api.get_instance_nw_info.return_value)
